from ontic import (core,
                   property,
                   schema,
                   type)
from ontic.property import OnticProperty
from ontic.schema import Schema
from ontic.type import OnticType

__all__ = [
    'core',
    'property',
    'schema',
    'type',
    'OnticProperty',
    'Schema',
    'OnticType',
]
